# Model Managers.

from django.db import models
from django.template import Template
from django.template import Context
import types

class XPathRenderer(object):
    """This is an interface for objects that have __xml__"""
    def __xml__(self, *args):
        return ""

def xmlify(qs, use_values=True, **kwargs):
    """XML serializer for queryset qs using the template described in kwargs.

    This operates a little like django values (indeed it uses django
    values if use_values is True) but allows you to specify the
    resulting XML name.

    So this call:

       xmlify(qs, **{
          "xml_field_name": "django_queryset__model__accessor",
          "other_xml_field": "django_queryset__deep__reference__field",
          "xml_field": "django_queryset__model__field|upper",
          }).__xml__()

    produces:

     [ "xml_field_name='value from django_queryset.model.accessor row 1'
         other_xml_field='value from django_queryset.deep.reference.field row 1'
         xml_field='value to caps from django_queryset.model.field row 1'",

       "xml_field_name='value from django_queryset.model.accessor row 2'
         other_xml_field='value from django_queryset.deep.reference.field row 2'
         xml_field='value to caps from django_queryset.model.field row 2'",
       .
       .
       .
       ]

    Any valid django template may be used on the django field name.

    If use_values is False then a normal queryset is used instead of a
    values queryset.

    The best way to use this with XSLT is to attach xmlify-ed
    querysets to a request context and then call render_to_response
    with the context object. 'xdjango:contextobject()' can then
    retrieve the queryset.
    """
    captured_qs = qs
    class XML(XPathRenderer):
        def __init__(self):
            self._cached = None

        def __xml__(self, *args):
            if self._cached == None:
                self._cached = self.__evalxml__(*args)
            return self._cached

        def __evalxml__(self, *args):
            template_list = [(name, Template('{{%s}}' % value)) \
                                 for name,value in kwargs.iteritems()]
            django_fields = [template.split("|")[0].split(".")[0] for template in kwargs.values()]

            # Do the query that gets the data to XML
            # Ordinarily we use 'values' but we can use an ordinary query if necessary
            rows = []
            text_fields = []
            if use_values:
                rows = captured_qs.values(*django_fields)
            else:
                for row in captured_qs:
                    row_result = {}
                    if django_fields:
                        for field in django_fields:
                            value = getattr(row, field)
                            row_result[field] = value() if isinstance(value, types.MethodType) else value
                            if getattr(value, 'is_text', False):
                                text_fields.append(field)
                    else:
                        row_result = row.__xml__()

                    rows += [row_result]

            # Make a nice list of template outputed rows
            from lxml import etree
            xmlname = captured_qs.model.__name__
            xmlroot = etree.Element("%ss" % xmlname.lower())
            #import pdb
            #pdb.set_trace()
            for record in rows:
                c = Context()
                c.update(record)
                child = etree.SubElement(xmlroot, xmlname.lower())
                if template_list:
                    for name,template in template_list:
                        if name in text_fields:
                            elem = etree.SubElement(child, name)
                            elem.text = template.render(c)
                        else:
                            child.attrib[name] = template.render(c)
                else:
                    parsed = etree.XML(record)
                    elem = child.append(parsed)

            return xmlroot
                
    return XML()

def xmlifyiter(iterator, name, **kwargs):
    """XML serializer to elements of 'name'.

    The kwargs specifies the XML -> iterator values mapping.  

    At the moment the iterator values are expected to be dictionary
    type objects.

    The kwargs specifies:

      an xml result attribute name = "an iterator value object dictionary key"

    for example:

      xmlifyiter(iteratorobject, "Iterable", name="username", age="age")

    =>
      <iterables>
         <iterable name="nicferrier" age="40"/>
         <iterable name="neebone" age="29"/>
         <iterable name="iamseb" age="31"/>
      </iterables>
    """
    captured_iter = iterator
    captured_element_name = name
    class XML(XPathRenderer):
        def __init__(self):
            self._cached = None

        def __xml__(self, *args):
            if self._cached == None:
                self._cached = self.__evalxml__(*args)
            return self._cached

        def __evalxml__(self, *args):
            template_list = [(name, Template('{{%s}}' % value)) \
                                 for name,value in kwargs.iteritems()]
            dict_keys = [template.split("|")[0].split(".")[0] for template in kwargs.values()]

            rows = []
            text_fields = []
            for row in captured_iter:
                row_result = {}
                for field in dict_keys:
                    value = row.get(field)
                    row_result[field] = value
                    if getattr(value, 'is_text', False):
                        text_fields.append(field)
                rows += [row_result]

            # Make a nice list of template outputed rows
            from lxml import etree
            xmlname = captured_element_name
            xmlroot = etree.Element("%ss" % xmlname.lower())
            #import pdb
            #pdb.set_trace()
            for record in rows:
                c = Context()
                c.update(record)
                child = etree.SubElement(xmlroot, xmlname.lower())
                for name,template in template_list:
                    if name in text_fields:
                        elem = etree.SubElement(child, name)
                        elem.text = template.render(c)
                    else:
                        child.attrib[name] = template.render(c)
            return xmlroot
                
    return XML()


from django.db.models.query import QuerySet
class XmlQuerySet(QuerySet):
    """A queryset that does xmlifying"""
    def xml(self, **kwargs):
        return xmlify(self, use_values=getattr(self, "use_values", True), **kwargs)

    def xml_objects(self, **kwargs):
        """Don't do a values query. Use the full object instead."""
        return xmlify(self, use_values=getattr(self, "use_values", False), **kwargs)


def monkey_qs(qs, use_values=True):
    """Clone the queryset adding an 'xml' method.

    The 'xml' method works like 'xmlify', pass kwargs for rendering
    arguments and it returns an object which supports the __xml__
    protocol.

    The objects returned from the 'xml' object attached with
    'monkey_qs' are also querysets which can be further cloned.

    For example:

      qs1 = Person.objects.filter(user__last_name="smith")
      qs2 = monkey_qs(qs1)
      qs3 = qs2.xml(username="user__username", firstname="user__firstname")
      qs4 = qs3.filter(age__gte=18)
      xmldata = qs4.__xml__()
    """
    capturedclone = qs._clone

    def baseclone(xml_closure):
        newclone = capturedclone()
        newclone.__xml__ = lambda *args: xml_closure.__xml__(*args)
        newclone._clone = lambda: baseclone(xml_closure)
        return newclone

    def adapt(*args, **kwargs):
        newclone = capturedclone(*args, **kwargs)
        xml_protocol_obj = xmlify(qs, use_values=use_values, **kwargs)
        newclone.__dict__["__xml__"] = lambda *args: xml_protocol_obj.__xml__(*args)
        newclone.__dict__["_clone"] = lambda: baseclone(xml_protocol_obj)
        return newclone

    def xmlclone(selfarg, *args, **kwargs):
        newclone = capturedclone(*args, **kwargs)
        newclone.__dict__["xml"] = lambda *args,**kwargs: adapt(*args, **kwargs)
        newclone.__dict__["use_values"] = use_values
        newclone.__dict__["_clone"] = types.MethodType(xmlclone, selfarg, selfarg.__class__)
        return newclone
    return xmlclone(qs)


class RenderingManager(models.Manager):
    """Use monkey_qs to decorate the queryset.

    This makes querysets that use values calls by default. To get a
    queryset that will render without using values create the manager
    with:

      use_values=False
    """
    def __init__(self, use_values=True):
        models.Manager.__init__(self)
        self.use_values = use_values

    def get_query_set(self):
        qs = super(RenderingManager, self).get_query_set()
        return monkey_qs(qs)

# End



"""Models for XSLT.

These are purely for testing the XSLT code.
"""

from django.db import models
from managers import RenderingManager

# A test manager
class XSLTTestManager(RenderingManager):
    pass

# A test model
class XSLTTestModel(models.Model):
    name = models.CharField(max_length=50)
    about = models.TextField()
    count = models.IntegerField()

    # Setup the manager to be a rendering manager
    objects = XSLTTestManager()

    def __xml__(self):
        return "<name>%s</name>" % self.name

# End



# encoding=utf-8

"""Tests for xslt"""

import time
import re
from django.template import Context
from testhelp import assertXpath
from unittest import TestCase
from djangoxslt import xslt
import logging
logging.basicConfig()

BLANK = """<?xml version="1.0" encoding="utf-8"?>
<xsl:stylesheet  version="1.0" 
                 xmlns="http://www.w3.org/1999/xhtml"
                 xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                 xmlns:xdjango="http://djangoproject.com/template/xslt"
                 extension-element-prefixes="xdjango"
                 exclude-result-prefixes="xdjango">
    <xsl:output omit-xml-declaration="yes"/>
    <xsl:template match="/">
%s
    </xsl:template>
</xsl:stylesheet>
"""

class DjangoContextFuncTest(TestCase):
    def setUp(self):
        super(DjangoContextFuncTest, self).setUp()
        self.func = xslt.DjangoContextFunc("testfunc")
        c = Context({'testfunc': 'a value'})
        xslt.djangothread.context = c

    def test_parse_str(self):
        result = self.func.parse("<tag>no</tag>")

    def test_parse_unicode(self):
        result = self.func.parse(u"<tag>sí</tag>")

    def test_parsehtml_str(self):
        result = self.func.parsehtml("<a>no</a>")

    def test_parsehtml_unicode(self):
        result = self.func.parsehtml(u"<a>sí</a>")

    def test_call_str(self):
        result = self.func("testval")
        assert result == "a value"

    def test_call_unicode(self):
        c = Context({'testfunc': u'☃☃☃☃☃'})
        xslt.djangothread.context = c
        result = self.func(u"téstv☃al")
        assert result == u'☃☃☃☃☃'

    def tearDown(self):
        xslt.djangothread.context = None


from django.test.client import Client

class XSLTTest(TestCase):
    """Test the xslt system..
    """
    def setUp(self):
        self.client = Client()

    def test_simple_page(self):
        """Test that we can retrieve the simple page."""
        response = self.client.get("/testtransform/simplepage/")
        self.assertEquals(response.status_code, 200)
        # We should really assert some xpath things about it.
        
from djangoxslt.xslt import managers as xsltmanagers

def _qs_eval_helper(qs):
    """This is a useful little bit of code to eval the xml dom result.

    The qs MUST have the __xml__ method, ie: it must result from an
    .xml(...) somethere in the chain."""

    from lxml import etree
    data1 = qs.__xml__()
    strresult = etree.tostring(data1)
    # print strresult
    return strresult

class IterableTest(TestCase):
    def setUp(self):
        super(IterableTest, self).setUp()
        self.time = int(time.time() * 1000)

    def test_iterable_render(self):
        tmpl = BLANK % """
        <xsl:copy-of select="xdjango:foo%d()"/>
        """ % self.time
        transformer = xslt.Transformer(tmpl)
        
        dictlist = [
            {"a": 10, "b": 20, "c": 40},
            {"a": 11, "b": 21, "c": 41},
            {"a": 12, "b": 22, "c": 42},
            ]

        xmlobj = xsltmanagers.xmlifyiter(
            dictlist,
            "Simple",
            attriba="a",
            otherattrib="c"
            )

        # Use objects and NOT values
        context_key = 'foo%d' % self.time
        c = Context({ 
                context_key: xmlobj,
                })

        res = transformer(context=c)
        assertXpath(
            res, 
            '//simples/simple[@attriba="%s"]' % 10
            )

class QSRenderTestCase(TestCase):
    def setUp(self):
        super(QSRenderTestCase, self).setUp()
        self.time = int(time.time() * 1000)

    def test_queryset_render_non_values(self):
        """Tests that querysets can be rendered without using 'values'.

        Takes a django User object and monkey_qs's it then test the
        xml is returned correctly when we use a dynamic method and
        therefore specify no to 'use_values'
        """
        tmpl = BLANK % """
        <xsl:copy-of select="xdjango:foo%d()"/>
        """ % self.time
        transformer = xslt.Transformer(tmpl)
        
        from django.contrib.auth.models import User
        user = User(
            username="user%s" % self.time,
            password="password%s" % self.time,
            first_name="first%s" % self.time,
            last_name="last%s" % self.time
            )
        user.save()

        # Use objects and NOT values
        qs = xsltmanagers.monkey_qs(
            User.objects.filter(username="user%s" % self.time),
            use_values=False
            )

        context_key = 'foo%d' % self.time
        xml = qs.xml(full_name="get_full_name", username="username")
        c = Context({ context_key: xml })
        res = transformer(context=c)
        assertXpath(
            res, 
            '//users/user[@full_name="first%s last%s"]' % (self.time, self.time)
            )

    def test_queryset_xmlify_user(self):
        """Tests that querysets can be rendered with the xmlify method.
        """
        tmpl = BLANK % """
        <xsl:copy-of select="xdjango:foo%d()"/>
        """ % self.time
        transformer = xslt.Transformer(tmpl)
        
        from django.contrib.auth.models import User
        user = User(
            username="user%s" % self.time,
            password="password%s" % self.time,
            first_name="first%s" % self.time
            )
        user.save()

        # Make a queryset
        qs = User.objects.filter(username="user%s" % self.time)

        # Make an xml object from it
        xml_list = xsltmanagers.xmlify(
            qs,
            first_name="first_name", 
            username="username"
            )

        context_key = 'foo%d' % self.time
        c = Context({ context_key: xml_list })
        res = transformer(context=c)
        assertXpath(
            res, 
            '//users//user[@first_name="first%s"]' % self.time,
            )


    def test_queryset_render_user(self):
        """Tests that querysets can be rendered.

        Takes a django User object and monkey_qs's it then test the
        xml is returned correctly.
        """
        tmpl = BLANK % """
        <xsl:copy-of select="xdjango:foo%d()"/>
        """ % self.time
        transformer = xslt.Transformer(tmpl)
        
        from django.contrib.auth.models import User
        user = User(
            username="user%s" % self.time,
            password="password%s" % self.time,
            first_name="first%s" % self.time
            )
        user.save()
        qs = xsltmanagers.monkey_qs(
            User.objects.filter(username="user%s" % self.time)
            )

        context_key = 'foo%d' % self.time
        xml = qs.xml(first_name="first_name", username="username")
        c = Context({ context_key: xml })
        res = transformer(context=c)
        assertXpath(
            res, 
            '//users/user[@first_name="first%s"]' % self.time,
            )

    def test_queryset_render_testmodel(self):
        """Tests that querysets can be rendered.

        Takes our own test model and uses the model's built in manager
        which monkey patches all querysets coming out of it.
        """
        tmpl = BLANK % """
        <xsl:copy-of select="xdjango:foo%d()"/>
        """ % self.time
        transformer = xslt.Transformer(tmpl)
        
        from models import XSLTTestModel
        testobject = XSLTTestModel(
            name = "name%s" % self.time,
            about = "about%s" % self.time,
            count = 10
            )
        testobject.save()

        # Do the query and pull back the xml
        xml = XSLTTestModel.objects.filter(
            name="name%s" % self.time
            ).xml(
            name="name", 
            about_text="about", 
            count="count"
            )
        c = Context({ 'foo%d' % self.time: xml })
        res = transformer(context=c)
        assertXpath(
            res, 
            '//xslttestmodels/xslttestmodel[@name="name%s"]' % self.time,
            )
        assertXpath(
            res, 
            '//xslttestmodels/xslttestmodel[@about_text="about%s"]' % self.time,
            )


    def test_queryset_render_testmodel_no_kwargs(self):
        """Tests that querysets can be rendered without kwargs.

        Takes our own test model and xmlify it with no kwargs causing
        it's own __xml__ method to be used to render the XML.
        """
        tmpl = BLANK % """
        <xsl:copy-of select="xdjango:foo%d()"/>
        """ % self.time
        transformer = xslt.Transformer(tmpl)
        
        from models import XSLTTestModel
        testobject = XSLTTestModel(
            name = "name%s" % self.time,
            about = "about%s" % self.time,
            count = 10
            )
        testobject.save()

        # Do the query and pull back the xml
        xml = XSLTTestModel.objects.filter(
            name="name%s" % self.time
            )

        xmlified = xsltmanagers.xmlify(xml, use_values=False)

        c = Context({ 'foo%d' % self.time: xmlified })
        res = transformer(context=c)
        assertXpath(
            res, 
            '//xslttestmodels/xslttestmodel/name[text()="name%s"]' % self.time,
            )



    def test_queryset_render_persists(self):
        """Tests that queryset __xml__ method attachment works.

        The __xml__ method, attached by using the xml() method, should
        'stick' to a queryset through further child-querysets.
        """
        tmpl = BLANK % """
        <xsl:copy-of select="xdjango:foo%d()"/>
        """ % self.time
        transformer = xslt.Transformer(tmpl)
        
        from models import XSLTTestModel
        for i in range(1,20):
            testobject = XSLTTestModel(
                name = "name%s" % (int(self.time) * i),
                about = "about%s" % (int(self.time) * i),
                count = i
                )
            testobject.save()

        # Do the query - this qs should have an 'xml' method
        base_qs = XSLTTestModel.objects.filter(
            name="name%s" % self.time
            )

        # Make a first sub-qs to check the xml method is being cloned
        qs = base_qs.filter(count__lte=12)

        # Check it has the xml method
        self.assert_("xml" in qs.__dict__)

        # Call the 'xml' method to store the xml to be generated and return a new qs
        qs_from_xml = qs.xml(name="name",  about_text="about", count="count")

        # Check it has the __xml__ method
        self.assert_("__xml__" in qs_from_xml.__dict__)

        # Make another qs from the 'xml' decorated one
        selected = qs_from_xml.filter(count__lte=3)

        # Check the sub-queryset has the __xml__ method
        self.assert_("__xml__" in selected.__dict__)

        xml_result = _qs_eval_helper(selected)
        self.assert_(re.search(""" name="name%s"[ /]""" % self.time, xml_result))
        
        c = Context({ 'foo%d' % self.time: selected })
        res = transformer(context=c)
        assertXpath(
            res, 
            '//xslttestmodels/xslttestmodel[@name="name%s"]' % self.time,
            )
        assertXpath(
            res, 
            '//xslttestmodels/xslttestmodel[@about_text="about%s"]' % self.time,
            )

        ##### TODO!!!!
        ### need to assert we have the RIGHT number of xslttestmodel objects
        ### it should be 12 and not 3 because we do the .xml(...) before the lte=3
    


class AVTTestCase(TestCase):
    def setUp(self):
        super(AVTTestCase, self).setUp()
        self.time = int(time.time() * 1000)

    def test_variable_usage(self):
        tmpl = BLANK % """
        <xsl:variable name="test" select="xdjango:foo%d()"/>
        <xsl:value-of select="$test"/>
        """ % self.time
        t = xslt.Transformer(tmpl)
        c = Context({'foo%d' % self.time: 'hello world'})
        assert t(context=c) == 'hello world\n'

    def test_simple(self):
        tmpl = BLANK % """
        <xsl:value-of select="xdjango:foo%d()"/>
        """ % self.time
        t = xslt.Transformer(tmpl)
        c = Context({'foo%d' % self.time: 'hello world'})
        assert t(context=c) == 'hello world\n'

    def test_rooted_avt(self):
        tmpl = BLANK % """
        <a href="{xdjango:foo%d()}/foo">!</a>
        """ % self.time
        t = xslt.Transformer(tmpl)
        c = Context({'foo%d' % self.time: 'some-location'})
        res = t(context=c)
        assertXpath(
            res, 
            '//xhtml:a[@href="some-location/foo"]',
            namespaces={
                "xhtml": "http://www.w3.org/1999/xhtml",
                }
            )

    def test_nonroot_avt(self):
        tmpl = BLANK % """
        <a href="/{xdjango:foo%d()}/foo">!</a>
        """ % self.time
        t = xslt.Transformer(tmpl)
        c = Context({'foo%d' % self.time: 'some-location'})
        res = t(context=c)
        assertXpath(
            res, 
            '//xhtml:a[@href="/some-location/foo"]',
            namespaces={
                "xhtml": "http://www.w3.org/1999/xhtml",
                }
            )

    def test_nonroot_avt_method(self):
        tmpl = BLANK % """
        <a href="/{xdjango:foo%d.upper()}/foo">!</a>
        """ % self.time
        t = xslt.Transformer(tmpl)
        c = Context({'foo%d' % self.time: 'some-location'})
        res = t(context=c)
        assertXpath(
            res, 
            '//xhtml:a[@href="/SOME-LOCATION/foo"]',
            namespaces={
                "xhtml": "http://www.w3.org/1999/xhtml",
                }
            )

    def test_method(self):
        tmpl = BLANK % """
        <xsl:value-of select="xdjango:foo%d.upper()"/>
        """ % self.time
        t = xslt.Transformer(tmpl)
        c = Context({'foo%d' % self.time: 'hello world'})
        assert t(context=c) == 'HELLO WORLD\n'

    def test_multiple_avt(self):
        tmpl = BLANK % """
        <a href="/{xdjango:foo%da()}/{xdjango:foo%db()}/foo">!</a>
        """ % (self.time, self.time)
        t = xslt.Transformer(tmpl)
        c = Context({
                'foo%da' % self.time: 'some-location',
                'foo%db' % self.time: 'more'
                })
        res = t(context=c)
        assertXpath(
            res, 
            '//xhtml:a[@href="/some-location/more/foo"]',
            namespaces={
                "xhtml": "http://www.w3.org/1999/xhtml",
                }
            )

# End



from django.http import HttpResponse
from django.template import RequestContext
from django.conf import settings
from os.path import join

import logging

from engine import TransformerFile
from engine import EMPTYDOC
from django.conf import settings

DEFAULT_PAGE_NAMESPACE=""        # WooMe's page namespace is "woome"
DEFAULT_PAGE_PATTERN="%s%s.xslt" # WooMe's page pattern is "%s_%s.xslt"

def page(request, page="index", namespace=DEFAULT_PAGE_NAMESPACE, **kwargs):
    """A generic XSLT view which just runs a page name derived XSLT file.

    Pass in a page to be rendered (this could come from a urls
    statement) and optionally a namespace.

    The 'namespace' allows you to namespace all the XSLT files that
    will be used like this in your transforms directory. For example,
    to serve a page ^main/$ you could write an XSLT called
    static_main.xslt. To use that you would declare your page url
    like:

      '(r'(?P<page>.*)/$', 'djangoxslt.xslt.views.page', {"namespace": "static"}),
      
    In order to make the namespace work we also allow a page pattern
    to be specified in the settings file. The page pattern is used to
    define how the namespace and the page name combine. The above
    example would require the following declared in settings.py:

       XSLT_PAGE_PATTERN="%s_%s.xslt"
    """
    logger = logging.getLogger("xslt.views.page")
    logger.info("page = %s namespace = %s" % (page, namespace))
    page_pattern = getattr(settings, "XSLT_PAGE_PATTERN", DEFAULT_PAGE_PATTERN)
    p = page_pattern % (namespace, page)
    t = TransformerFile(join(settings.TRANSFORMS, p))
    c = RequestContext(request, {})
    c.update(kwargs)
    out = t(EMPTYDOC, context=c)
    return HttpResponse(out)


# End



# XSLT helper
from __future__ import with_statement

"""
A Django template engine for XSLT.

This isn't a real template engine, in the sense that it does not
provide any Template class. We'll work towards that.

It does allow Django to use XSLT templates and to include context
variables in your templates with an xpath syntax.

In order to do this we implement an extension system that makes
context variables map to xpath functions. 

A context variable called as a function with no arguments returns the
string rendering of the context variable.

However, a context variable called with an argument is intended to
allow abstracted rendering behaviour. The argument is called the
''mapper key'' and provides a string -> callable mapping which will be
used to render the of the context variable to a legal XPath value. The
callable mapped to is called the ''renderer''.

Three renderers are defined by default:

 * xml 

   returns the context variable string evaluated but wrapped in a DIV
   element.

 * parse

   attempts to XML parse the context variable string evaluation.

 * parsehtml

   attempts to HTML parse the context variable string evaluation, the
   parsed document is normalized to XML before being returned so that
   it can be succesfully evaluated by XSLT.

   Currently, we fix any HTML parsed document with the XHTML namespace.


Further renderers may be specified in settings with the variable
XSLT_MAPPER:

 import project.render
 XSLT_MAPPER = {
       "project_special_render": project.render.do_render
    }

mapping keys defined in settings may override the default renderers.
"""

from django.conf import settings

from lxml import etree
from lxml.builder import E
from StringIO import StringIO
import re
import logging
import traceback

# This is a simple empty document you can pass into Transformer.__call__ if you need to.
EMPTYDOC = etree.Element("empty")

# DO NOT change these without running the xslt unit-tests in this module.
FUNC_MATCH_RE1 = re.compile(r"{?xdjango:[^}]+}?")
FUNC_MATCH_RE2 = re.compile(r"{?xdjango:([^(]+)\((.*?)\)}?")

# taken from drivel.config
def dotted_import(name):
    mod, attr = name.split('.'), []
    obj = None
    while mod:
        try:
            obj = __import__('.'.join(mod), {}, {}, [''])
        except ImportError, e:
            attr.insert(0, mod.pop())
        else:
            for a in attr:
                try:
                    obj = getattr(obj, a)
                except AttributeError, e:
                    raise AttributeError('could not get attribute %s from %s -> %s (%r)' % (
                        a, '.'.join(mod), '.'.join(attr), obj))
            return obj
    raise ImportError('could not import %s' % name)


class Exml(object):
    """A simple document builder for lxml"""
    def __init__(self, element_name, **attribs):
        self.el = None
        self.element_name = element_name
        self.attribs = attribs

    def append(self, element_name, **attribs):
        if self.el == None:
            self.el = etree.Element(self.element_name)
            for name,value in self.attribs.iteritems():
                self.el.set(name,str(value))

        ne=etree.SubElement(self.el, element_name)
        for name,value in attribs.iteritems():
            ne.set(name,str(value))

        ec = Exml(element_name, **attribs)
        ec.el = ne
        return ec


class Xml(object):
    def __init__(self, data, base_url=None, parser=None):
        self.parser = parser if parser else etree.XMLParser()
        self.data = data
        self.base_url = base_url

    def __call__(self):
        doc = etree.fromstring(
            self.data,
            parser=self.parser,
            base_url=self.base_url)
        return doc

class Html(Xml):
    def __init__(self, data, base_url=None, parser=None):
        super(Html,self).__init__(data, base_url, parser if parser else etree.HTMLParser())

XHTML_NAMESPACE = "http://www.w3.org/1999/xhtml"
DJANGO_NAMESPACE="http://djangoproject.com/template/xslt"

from django.template import Variable
from django.template import VariableDoesNotExist
import types
import threading
djangothread = threading.local()

class DjangoContextFunc(object):
    """Implements an extension function for django contexts"""
    def __init__(self, name, context=None):
        self.logger = logging.getLogger("xslt.DjangoContextFunc.%s" % name)
        self.logger.debug("creating %s" % name)
        self.name = name
        # Define some default mappers and extend with settings
        self.mappers = {
            "xml": self.xml,
            "parsehtml": self.parsehtml,
            "parse": self.parse
            }
        self._context = context
        try:
            self.mappers.update(settings.XSLT_MAPPER)
        except AttributeError:
            pass

    @property
    def context(self):
        if self._context is None:
            return djangothread.context
        return self._context

    def _django_eval(self):
        """
        When the variable doesn't resolve, we return an empty string,
        this allows for tests such as:
        <xsl:if test="django:form.errors()">
            <xsl:value-of select="django:form.errors.fieldname" />
        </xsl:if>
        
        Where form.errors is a dict, and form.errors['fieldname'] 
        may not exist, returning an empty string will allow evaluation
        to continue.
        """
        try:
            e = Variable(self.name).resolve(self.context)
            return e
        except VariableDoesNotExist, e:
            # Nic says: I think there should be some debug setting
            # here to make this report the error
            errmsg = "problem evaling variable %s %s %s" % (
                self.name,
                e.__class__.__name__,
                e
                )
            self.logger.error(errmsg)
            return ""
        except Exception, e:
            errmsg = "problem evaling variable %s %s %s" % (
                self.name,
                e.__class__.__name__,
                e
                )
            self.logger.error(errmsg)
            if settings.DEBUG:
                return [E.error(errmsg)]
            return ""

    def parsehtml(self, ctx_value, *args):
        try:
            # First make it HTML
            htmldoc = etree.HTML(ctx_value)
            xmldoc = etree.XML(re.sub(
                    "<html>", 
                    """<html xmlns="%s">""" % XHTML_NAMESPACE,
                    etree.tostring(htmldoc)
                    ))
            self.logger.debug(etree.tostring(xmldoc))
            doc= [xmldoc]
            return doc
        except etree.XMLSyntaxError, e:
            self.logger.debug(ctx_value)
            for i in e.error_log:
                self.logger.error("couldn't transform %s" % i)
                self.logger.debug(traceback.format_exc())
            if settings.DEBUG:
                errors = [E.li(str(error)) for error in e.error_log]
                return [E.ol(*errors)]
            return ""

    def parse(self, ctx_value, *args):
        try:
            ## Not sure if it's better to return EMPTYDOC from here if nothing is passed in
            if ctx_value:
                xmldoc = etree.fromstring(unicode(ctx_value))
                doc= [xmldoc]
                return doc
            else:
                return EMPTYDOC
        except etree.XMLSyntaxError, e:
            self.logger.debug(ctx_value)
            for i in e.error_log:
                self.logger.error("couldn't transform %s via %s %s" % (i, ctx_value, args))
                self.logger.debug(traceback.format_exc())
            if settings.DEBUG:
                errors = [E.li(str(error)) for error in e.error_log]
                return [E.ol(*errors)]
            return ""

    def xml(self, ctx_value, *args):
        e = etree.Element("div" if len(args) < 2 else args[1])
        e.text = ctx_value
        return [e]
    
    def __call__(self, ctx, *args):
        """Treat a django context variable as an XSLT callable.

        If the context object supports the __xml__ protocol then the
        method is called (with any arguments) and the return value is
        expected to be some XML value like an lxml DOM or an XPath
        value, like a string.

        If the call specifies a renderer name as arg #1 then the
        renderer name is looked up in settings.XSLT_MAPPER and the
        resulting callable name evaled and then passed the context
        variable and the rest of the args.
        
        If the context object does not support __xml__ and the call
        does not specify a renderer then it is just str(evaled).

        The arguments to this function are the normal xpath extension
        function arguments: 'ctx' is an opaque xpath context value and
        'args' is (mostly) the list of arguments supplied to the xpath
        function.
        """
        self.logger.debug("called %s" % args[0] if len(args) > 0 else "")
        try:
            ctx_value = self._django_eval()
            self.logger.debug("context value %s" % ctx_value)

            # If the context object supports the render protocol use it
            try:
                # We just EXPECT the value to be XML
                if len(args):
                    return ctx_value.__xml__(*args)
                return ctx_value.__xml__()
            except AttributeError:
                pass
            
            # If there are no args there is no defined renderer
            if not len(args):
                if isinstance(ctx_value, basestring):
                    return ctx_value
                return str(ctx_value)

            # If the argument are xpath evaled they are general complex
            # lxml returns them as lists... python is expecting something simpler
            # not sure the best way to do this... maybe a higher level API
            # passing a context to the function.
            a = [str(x[0]) if isinstance(x,type([])) else x \
                     for x in args]
            fn = a.pop(0)

            # This is a useful place to put trace breaks
            # you can test fn for your renderer name
            renderer_mapped = self.mappers[fn]
            if isinstance(renderer_mapped, basestring):
                try:
                    renderer = renderer_mapped.split(".")
                    self.logger.debug("found renderer %s" % ".".join(renderer))
                    r = dotted_import(renderer_mapped)
                    value = r(ctx_value, *a)
                except Exception, e:
                    errmsg = "renderer %s had error %s" % (
                        ".".join(renderer),
                        e)
                    self.logger.error(errmsg, exc_info=True)
                    return [E.error(errmsg)]
                else:
                    return value
            elif isinstance(renderer_mapped, types.FunctionType):
                value = renderer_mapped(ctx_value, *a)
                self.logger.debug("%s returning %s", fn, value)
                return value
            elif isinstance(renderer_mapped, types.MethodType):
                value = renderer_mapped(ctx_value, *a)
                self.logger.debug("%s returning %s", fn, value)
                return value
        except Exception, e:
            errmsg = "resolving context call %s had error %s %s" % (
                self.name,
                e.__class__.__name__,
                e)
            self.logger.error(errmsg, exc_info=True)
            return [E.error(errmsg + traceback.format_exc())]
            
        
class DjangoResolver(etree.Resolver):
    """A base django resolver.

    This is necessary to discover functions used in  every XSLT document.

    The resolver connects xpath expressions of the kind:

      xdjango:contextvarname(args)

    to context variable mapping implementations; eg:

      xdjango:contextvar('xml')

    will cause the Django context variable 'contextvar' to be rendered 
    with the XML context variable renderer, while:

      xdjango:contextvar('reported')

    will cause the Django context variable 'contextvar' to be rendered
    with the renderer 'reported', presumably a custom renderer for that
    data type.
    """

    def __init__(self):
        self.parser = etree.XMLParser()

    def _resolve(self, content, context, base_url=None):
        xml = etree.parse(StringIO(content), self.parser)

        # We should check this is an xslt document
        results = xml.xpath(
            "//@*",
            namespaces={
                # The tag used here is relied upon in the following regexing
                "xdjango": DJANGO_NAMESPACE,
                })
        # Not a perfect regex here, {} should wrap, or not.
        djangocalls = [r for r in results if FUNC_MATCH_RE1.search(r)]

        fns = etree.FunctionNamespace(DJANGO_NAMESPACE)
        for call in djangocalls:
            offset = 0
            while True:
                m = FUNC_MATCH_RE2.search(call, offset)
                if m is None:
                    break
                name = m.group(1)
                offset = m.end()
                if name not in fns:
                    fns[name] = DjangoContextFunc(name)
        # We want to call the actual super here
        return super(DjangoResolver, self).resolve_string(
            content, 
            context, 
            base_url=base_url)

    def resolve(self, url, pubid, context):
        """Implement resolver for django transformers.

        At the moment this seems to only be able to do file based resolving.
        That's because I'm not sure how to differentiate on the url.
        """
        if url.partition(':')[0] not in ['http', 'django', 'querydirect']:
            # FIXME
            # We need a decent error here to say we couldn't find it.
            with open(url) as fd:
                content = fd.read()
                return self._resolve(content, context, base_url=url)

    def resolve_file(self, f, context, base_url=None):
        content = f.read()
        return self._resolve(content, context, base_url=base_url)

    def resolve_filename(self, filename, context):
        with open(filename) as f:
            content = f.read()
            return self._resolve(content, context, base_url=filename)
    
    def resolve_string(self, content, context, base_url=None):
        return self._resolve(content, context, base_url=base_url)


# Hook management

_transformer_init_hook_list = []
_transformer_percall_hook_list = []

def _transformer_init_hook(transformer_object):
    """Purely backward stuff
    
    This links our two xslt systems.
    """
    global _transformer_init_hook_list
    for hook_func in _transformer_init_hook_list:
        hook_func(transformer_object)

def _transformer_percall_hook(transformer_object, doc, context, **params):
    for hook_func in _transformer_percall_hook_list:
        hook_func(transformer_object, doc, context, **params)

def add_init_hook(hookfunc):
    """Add the specified function to the list of functions called when we init a transformer."""
    global _transformer_init_hook_list
    if hookfunc not in _transformer_init_hook_list:
        _transformer_init_hook_list += [hookfunc]

def add_percall_hook(hookfunc):
    """Add the specified function to the list of functions called when we call a transformer.

    The hook must be a function like this:

      function(transformer_object, doc, context, **params)
    
    the parameters should be obvious.
    """
    global _transformer_percall_hook_list
    if hookfunc not in _transformer_init_hook_list:
        _transformer_percall_hook_list += [hookfunc]


# Transformers

class Transformer(object):
    def __init__(self, 
                 content, 
                 resolv=lambda c,p: etree.fromstring(c,p),
                 parser=None,
                 context=None):
        """Make a transformer object.

        The transformer wraps all the django specific functionality.
        
        Params:
          content is the XML content, this could be a string or a url
             the exact semantics of content are defined by resolver
           
          resolv is a function that is called to return the XSLT
          document, it is called like this:

              xsltdoc = resolv(content, parser)

          where content and parser are both the args from __init__.
          The default value of resolv is:

              lambda c,p: etree.fromstring(c,p)

          parser is the XMLParser to use. a default is supplied. 
        """
        context = context if context else {}

        global DJANGO_NAMESPACE
        self.logger = logging.getLogger("xslt.Transformer")
        fns = etree.FunctionNamespace(DJANGO_NAMESPACE)

        # Setup the rest of the environment
        self.parser = parser if parser else etree.XMLParser()

        # We call out here to anything that's defined
        _transformer_init_hook(self)

        # Setup the djangoxslt resolver
        self.resolver = DjangoResolver()
        self.parser.resolvers.add(self.resolver)

        # lxml doesn't seem to use the parser's resolver for this
        # Hence we need to do the great big hack below
        self.xslt_doc = resolv(content, self.parser)

        ## Great big hack
        # We should check this is an xslt document
        xml = self.xslt_doc
        results = xml.xpath(
            "//@*",
            namespaces={
                # The tag used here is relied upon in the following regexing
                "xdjango": DJANGO_NAMESPACE,
                })
        # Not a perfect regex here, {} should wrap, or not.
        djangocalls = [r for r in results if FUNC_MATCH_RE1.search(r)]
        for call in djangocalls:
            offset = 0
            while True:
                m = FUNC_MATCH_RE2.search(call, offset)
                if m is None:
                    break
                name = m.group(1)
                offset = m.end()
                if name not in fns:
                    fns[name] = DjangoContextFunc(name)
        # End Great big hack

        qs_extension = QuerySetTemplateElement()
        extensions = {(DJANGO_NAMESPACE, 'queryset'): qs_extension}
        self.xslt = etree.XSLT(self.xslt_doc, extensions=extensions)

    def __xslt_error__(self, errorlist):
        """Format an errorlist.

        Override this if you want your errors looking different.
        """
        errors = [E.li(str(error)) for error in errorlist]
        errordoc = E.html(
            E.h1("an error occurred"),
            E.ol(*errors)
            )
        return errordoc

    def __call__(self, 
                 doc=None, 
                 context=None, 
                 **params):
        from django.template import Context
        global djangothread
        djangothread.context = context if context != None else Context()
        doc = doc if doc is not None else EMPTYDOC

        # Call out to the percall hooks
        _transformer_percall_hook(self, doc, context, **params)

        # accept a string for document as well
        if isinstance(doc, basestring):
            doc = etree.fromstring(doc)

        try:
            return str(self.xslt(doc, **params))
        except etree.XSLTApplyError, e:
            self.logger.error("couldn't transform %s" % e.error_log)
            self.logger.error("couldn't transform %s" % e)
            self.logger.debug(self.__xslt_error__(e.error_log))

            if settings.DEBUG:
                return etree.tostring(self.__xslt_error__(e.error_log))
            else:
                raise
        except etree.XMLSyntaxError, e:
            for i in e.error_log:
                self.logger.error("couldn't transform %s" % i)
                self.logger.debug(traceback.format_exc())
            if settings.DEBUG:
                return etree.tostring(self.__xslt_error__(e.error_log))
            else:
                raise


from os.path import join as joinpath

def transformer_file_resolv_callback(c, p):
    """A resolver function for Transformer.

    c is the content which should be a filename
    p is the parser which we'll use.
    """
    logger = logging.getLogger("djangoxslt.xslt.transformer_file_resolv_callback")
    try:
        stylesheet = joinpath(*c)
        return etree.parse(stylesheet, p)
    except etree.XMLSyntaxError:
        logger.error("stylesheet %s" % stylesheet)
        raise

class TransformerFile(Transformer):
    """Make a transformer from an XSLT file.

    You can pass through a list of filename parts that will be joined
    to construct a filename."""

    def __init__(self, *filename_parts, **kwargs):
        try:
            stylesheet = joinpath(filename_parts)
            self.stylesheet = stylesheet
            super(TransformerFile, self).__init__(
                stylesheet, 
                resolv=transformer_file_resolv_callback,
                **kwargs)
        except Exception, e:
            e.stylesheet = stylesheet 
            e.message = "%s {%s}" % (e.message, e.stylesheet)
            raise

from django.http import HttpResponse
def render_to_response(xslt, context, mimetype="text/html"):
    t = TransformerFile(settings.TRANSFORMS, xslt)
    return HttpResponse(t(context=context), mimetype="text/html")

class QuerySetTemplateElement(etree.XSLTExtension):
    def execute(self, context, self_node, input_node, output_parent):
        ctx = djangothread.context
        key = self_node.get('key')
        dest = self_node.get('dest')
        if '.' in key:
            qs = DjangoContextFunc(key, context=ctx)(None, 'pass')
        else:
            qs = ctx[key]
        for item in qs:
            ctx[dest] = item
            el = etree.Element('{%s}%s' % (DJANGO_NAMESPACE, dest))
            #self.apply_templates(context, el, output_parent)
            results = self.apply_templates(context, el)
            content = results[0]
            if isinstance(content, basestring):
                output_parent.text = content
            else:
                output_parent.append(content)
            #try:
                #self.process_children(context, output_parent)
            #except AttributeError, e:
                #raise RuntimeError('template not specified')
            del ctx[dest]

# End



#from setuptools import setup
#from setuptools import find_packages
from setuptools import setup

classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Utilities',
    'Topic :: Communications :: Email',
]

# Depends: 
# django
setup(
    name = "django-xslt",
    version = "0.4.4",
    description = "an XSLT template system for Django",
    long_description = """A replacment for Django's template system based on XSLT.""",
    license = "BSD",
    author = "Nic Ferrier",
    author_email = "nic@woome.com",
    url = "http://github.com/woome/django-xslt",
    download_url="http://github.com/woome/django-xslt/downloads",
    platforms = ["unix"],
    packages = ["djangoxslt", "djangoxslt.xslt"],
    package_dir = {"":"src"},
# Not sure we need a script, it would be nice to ship a django command line xsltproc?
#    scripts=['src/md'],   
    classifiers =  classifiers
    )



from lxml import etree
from StringIO import StringIO

# This is used as a parsed-XML cache
# We put all the content we parse here keyed by content
XMLCACHE={}

def assertXpath(xml, xpr, assertion_message="", namespaces=None, html=False):
    """Assert the Xpath 'xpr' evals against the 'xml' document.

    assertion_message can be specified as an alternative error message from a failure.
    namespaces can be specified as a list of namespace key:url pairs to be passed to XSLT
    html is boolean to specify whether to parse the document as HTML or not.
    """

    namespaces = {} if not namespaces else namespaces
    try:
        doc = XMLCACHE.get(xml)
        if not doc:
            parser = etree.HTMLParser() if html else etree.XMLParser()
            doc = etree.parse(StringIO(xml), parser)
            XMLCACHE[xml] = doc
    except Exception, e:
        raise e
    else:
        ret = doc.xpath(xpr, namespaces=namespaces) if namespaces else doc.xpath(xpr)
        if not ret:
            assertion_message = assertion_message \
                if assertion_message \
                else "{%s} did not evaluate with the specified document" % xpr
            raise AssertionError(assertion_message)

# End        
